
#include "TrackView.h"
#include <math.h>
#include <cstring>

#include "ScAnimation.h"
#include "tinyxml.h"
#include "ScException.h"
#include <Fl/Fl_JPEG_Image.h>
#include <Fl/Fl_PNG_Image.h>
#include <Fl/Fl_BMP_Image.h>
#include <Fl/Fl_PNM_Image.h>
#include <Fl/Fl_GIF_Image.h>
#include <Fl/Fl_XPM_Image.h>
#include <FL/Fl_File_Chooser.H>   // fl_alert
#include "GameConstants.h"
#include "TrackEditorUI.h"

#ifdef XCODE
#include <Carbon/Carbon.h>
#endif

#ifndef strtok_r
   #define strtok_r(s,d,p) strtok(s,d)
#endif

const unsigned int kMinimapUpdateLimit = 10;
const int kMainWindowWidth = 1024;
const int kMainWindowHeight = 768;
const int kScrollbarMaxValue = 10000;

void freeSearchDirectories(SearchDirectories *dirs) {
   SearchDirectories *temp;
   while (dirs != NULL) {
      temp = dirs->next;
      free(dirs);
      dirs = temp;
   }
}

SearchDirectories *makeSearchDirectories(const char *path, SearchDirectories *next) {
   SearchDirectories *result = new SearchDirectories;
   strncpy(result->path, path, 255);
   result->next = next;
   return result;
}

TiXmlDocument *findXMLFile(const char *filename, SearchDirectories *dirs, char *outPath = NULL)
{
   const int filenameLen = strlen(filename);
   char str[256];
   TiXmlDocument *xmlDoc;
   while (dirs != NULL) {
      if (strlen(dirs->path) + filenameLen + 1 < 255) {
         strcpy(str, dirs->path);
         strcat(str, "/");
         strcat(str, filename);
         xmlDoc = new TiXmlDocument(str);
         if (xmlDoc->LoadFile()) {
            if (outPath)
               strcpy(outPath, dirs->path);
            return xmlDoc;
         }
         else {
            delete xmlDoc;
         }
         dirs = dirs->next;
      }
   }
   return NULL;
}

//{jpg,png,bmp,pnm,pbm,pgm,ppm}
Fl_Image* loadImage(const char *filename)
{
   char *suffix = strrchr(filename, '.');
   if (strcmp(suffix,".jpg") == 0)
      return new Fl_JPEG_Image(filename);
   else if (strcmp(suffix,".png") == 0)
      return new Fl_PNG_Image(filename);
   else if (strcmp(suffix,".bmp") == 0)
      return new Fl_BMP_Image(filename);
   else if (strcmp(suffix,".pnm") == 0)
      return new Fl_PNM_Image(filename);
   else if (strcmp(suffix,".pbm") == 0)
      return new Fl_PNM_Image(filename);
   else if (strcmp(suffix,".pgm") == 0)
      return new Fl_PNM_Image(filename);
   else if (strcmp(suffix,".ppm") == 0)
      return new Fl_PNM_Image(filename);
   else if (strcmp(suffix,".gif") == 0)
      return new Fl_GIF_Image(filename);
   else if (strcmp(suffix,".xpm") == 0)
      return new Fl_XPM_Image(filename);
   else
      return NULL;
}

void PixelDataForImage(const Fl_Image* src, ScPixelData& dst)
{
	dst.bytes_pp = src->d();
	dst.width = src->w();
	dst.height = src->h();
	dst.pitch = dst.width * dst.bytes_pp;
	
	switch (dst.bytes_pp)
	{
		case 4:
			dst.format = GL_RGBA;
			dst.type = GL_UNSIGNED_BYTE;
			dst.internalformat = GL_RGBA;
			break;
		
		case 3:
			dst.format = GL_RGB;
			dst.type = GL_UNSIGNED_BYTE;
			dst.internalformat = GL_RGB;
			break;
		
		default:
			ScThrowErr("Only 32 bit or 24 bit is supported at present");
			break;
	}
	
   if (src->count() == 1)
      dst.pixels = src->data()[0];
   else
      ScThrowErr("Unsupported image type");
}

TrackView::TrackView(int x,int y,int w,int h,const char *l)
   : Fl_Gl_Window(x,y,w,h,l), startLine(0), view(0), searchPath(0), tiles(0)
{
#ifdef WIN32
   strncpy(dataDir, DATA_DIR, 1023);
   strncpy(userDataDir, DATA_DIR, 1023);
   searchPath = makeSearchDirectories(dataDir, NULL);
#else
 #ifdef XCODE
   // in the Xcode bundle DATA_DIR is only a relative directory so we need to make it absolute here.
   Boolean succ = CFURLGetFileSystemRepresentation(CFBundleCopyBundleURL(CFBundleGetMainBundle()), TRUE, (UInt8*)dataDir, 1023);
   //getcwd(dataDir, 1023 - strlen(DATA_DIR) - 1);
   if (!succ) {
      fprintf(stderr, "Error: Failed to get application bundle path\n");
      exit(1);
   }
   char *s = strrchr(dataDir, '/');
   s[1] = '\0';
   strcat(dataDir, DATA_DIR);
 #else
   strncpy(dataDir, DATA_DIR, 1023);
 #endif // XCODE
   strcat(dataDir,"/data");
   sprintf(userDataDir, "%s/.toycars", getenv("HOME"));
   searchPath = makeSearchDirectories(dataDir, makeSearchDirectories(userDataDir, NULL));
#endif // WIN32
   printf("dataDir = %s\nuserDataDir = %s\n", dataDir, userDataDir);
   
	TiXmlDocument *xmlDoc = findXMLFile("tilesets/default/default.xml", searchPath);
   if (!xmlDoc) {
      fl_alert("Could not find default tileset. Make sure we are in the same directory as the toycars app bundle.\n");
      exit(1);
   }
   delete xmlDoc;
}

TrackView::~TrackView()
{
   if (view)
      delete view;
   if (searchPath)
      freeSearchDirectories(searchPath);
}

ScTilemap* TrackView::loadTileset(const char* tilesetFilename, short border, short seperation, bool blend)
{
	char path[1024] = "tilesets/";
	char filename[1024];
   char foundPath[1024];
	
	strcat(path, tilesetFilename);
	strcat(path, "/");
	strcpy(filename, path);
	strcat(filename, tilesetFilename);
	strcat(filename, ".xml");
	
	// Load the xml file
	TiXmlDocument *xmlDoc = findXMLFile(filename, searchPath, foundPath);
	if (!xmlDoc)
	{
		ScThrowErr("xml file not found");
	}
   strcat(foundPath, "/");
   
	// Do the document parsing here
	TiXmlElement *xTileset = xmlDoc->FirstChildElement("tileset");
	if (!xTileset)
	{
		xmlDoc->Clear();
		delete xmlDoc;
		ScThrowErr("xml tileset element not found");
	}
	
	TiXmlHandle tilesetHandle(xTileset);
	
	TiXmlText *xImage = tilesetHandle.FirstChild("image").FirstChild().ToText();
	if (!xImage)
	{
		xmlDoc->Clear();
		delete xmlDoc;
		ScThrowErr("xml image element not found");
	}
	
	TiXmlText *xTilewidth = tilesetHandle.FirstChild("tilewidth").FirstChild().ToText();
	if (!xTilewidth)
	{
		xmlDoc->Clear();
		delete xmlDoc;
		ScThrowErr("xml tilewidth element not found");
	}
	
	TiXmlText *xTileheight = tilesetHandle.FirstChild("tileheight").FirstChild().ToText();
	if (!xTileheight)
	{
		xmlDoc->Clear();
		delete xmlDoc;
		ScThrowErr("xml tileheight element not found");
	}
	
	ScPixelData pixelData;
	Fl_Image* img;
	ScTilemap* tilemap;
	int tileWidth = strtol(xTilewidth->Value(), (char **)NULL, 10);
	int tileHeight = strtol(xTileheight->Value(), (char **)NULL, 10);
	
	if (tileWidth <= 0)
		ScThrowErr("tilewidth is invalid string");
	if (tileHeight <= 0)
		ScThrowErr("tileheight is invalid string");
	
	// create the tile map
	strcpy(filename, foundPath);
	strcat(filename, path);
	strcat(filename, xImage->Value());
	img = loadImage(filename);
	if (img == NULL)
		ScThrowErr("Could not load tiles image file");
	PixelDataForImage(img, pixelData);
	tilemap = new ScTilemap(pixelData, tileWidth, tileWidth, border, seperation, blend);
	delete img;
	
	printf("successfully created tileset from %s\n", filename);
	
	xmlDoc->Clear();
   delete xmlDoc;
	return tilemap;
}

void TrackView::resetScroll()
{
   view->scrollTo(0,0);
   boundScroll();
   updateScrollBars();
   redraw();
}

void TrackView::newMap(int w, int h)
{
	tiles->newMap(w, h, kWallTile);
   pathPolygon.clear();
   redraw();
   ui->setChanged(false);
   resetScroll();
}

void readPointsFromString(const char* str, Tuple* points, short& count)
{
    char *word, *last;
    char *temp = strdup(str);
    
    for (   word = strtok_r(temp, ";", &last), count = 0;
            word;
            word = strtok_r(NULL, ";", &last), count++)
    {
        points[count].x = atof(word);
        if ((word = strchr(word, ',')) == NULL)
            ScThrowErr("Badly formatted points list");
        points[count].y = atof(word + 1);
    }
    
    free(temp);
}

void TrackView::loadMap(const char *fullFilename)
{
	char filename[1024];
   char *temp;
   
	// Load the xml file
	TiXmlDocument *xmlDoc = new TiXmlDocument(fullFilename);
	if (!xmlDoc->LoadFile())
	{
		delete xmlDoc;
		ScThrowErr("xml file not found");
	}
	
	TiXmlHandle docHandle(xmlDoc);
	
	TiXmlElement *xTrack = docHandle.FirstChildElement("track").Element();
	TiXmlText *xTileset = docHandle.FirstChildElement("track").FirstChild("tileset").FirstChild().ToText();
	TiXmlText *xTilemap = docHandle.FirstChildElement("track").FirstChild("tilemap").FirstChild().ToText();
	TiXmlElement *xStartline = docHandle.FirstChildElement("track").FirstChildElement("startline").Element();
	TiXmlText *xPath = docHandle.FirstChildElement("track").FirstChild("path").FirstChild().ToText();
	
	// do error checking
	if (!(xTrack && xTileset && xTilemap && xStartline))
	{
		xmlDoc->Clear();
		delete xmlDoc;
		if (!xTrack)
			ScThrowErr("xml track element not found");
		if (!xTileset)
			ScThrowErr("xml tileset element not found");
		if (!xTilemap)
			ScThrowErr("xml tilemap element not found");
		if (!xStartline)
			ScThrowErr("xml startline element not found");
		if (!xPath)
			ScThrowErr("xml path element not found");
	}
	
	// load the tilemap from the specified tilemap file
	strncpy(filename, fullFilename, 1023);
	temp = (char *)strrchr(filename, '/');
	if (!temp)
		ScThrowErr("xml filename is not absolute");
   temp[1] = '\0';
	strcat(filename, xTilemap->Value());
	try { tiles->loadMap(filename); }
	catch (ScException e) {
      e.printMsg();
      newMap(80,50);
      return;
   }
   
	// the path
	pathPolygon.clear();
	if (xPath != NULL) {
		Tuple points[16384];
		short pointsCount;
		readPointsFromString(xPath->Value(), points, pointsCount);
		for (short k = 0; k < pointsCount; k++) {
			pathPolygon.push_back(points[k] * tiles->getTileWidth() / kScale);
		}
	}
	
	// get the start line attributes
	Tuple p1, p2;
	if (!(xStartline->Attribute("x1", &p1.x) && xStartline->Attribute("y1", &p1.y) &&
			xStartline->Attribute("x2", &p2.x) && xStartline->Attribute("y2", &p2.y)))
	{
		xmlDoc->Clear();
		delete xmlDoc;
		ScThrowErr("xml startline attributes not set");
	}
	p1.x *= tiles->getTileWidth();
	p2.x *= tiles->getTileWidth();
	p1.y *= tiles->getTileHeight();
	p2.y *= tiles->getTileHeight();
	
	startLine->setLine(p1, p2);
	
	// Clear and delete our allocated document and return success ;)
	xmlDoc->Clear();
	delete xmlDoc;
	
	printf("loading done\n");
   redraw();
   ui->setChanged(false);
   resetScroll();
}

void TrackView::init()
{
   try {
	ScTilemap* tilemap = loadTileset("default");	// will get deleted when animation destructor is called
	
	animation.push_back(tilemap);
	tiles = tilemap;
   printf("Current window dims: %d %d\n", this->w(), this->h());
	ScPortRect portRect = {0, 0, this->w(), this->h()};
	ScRect viewRect = {0, 0, portRect.width, portRect.height};
	view = new ScView(portRect, viewRect, animation);
	tilemap->newMap(80, 50, kWallTile);
	
	//list<Player*> emptyPlayersList;
	//minimap = new Minimap(*tilemap, emptyPlayersList);
	
	animation.push_back(new PathLine(tilemap->getBounds(), &pathPolygon));
	
	startLine = new Startline(Tuple(viewRect.right*0.5,viewRect.top*0.75+viewRect.bottom*0.25),
										 Tuple(viewRect.right*0.5,viewRect.top*0.25+viewRect.bottom*0.75),
										 kDefaultStartLineWidth, true);
	animation.push_back(startLine);
   
   updateScrollBars();
   }
   catch (ScException e) {
      printf("Failed to initialise. Make sure we are in the same directory as the toycars app bundle.\n");
      e.printMsg();
      exit(1);
   }
}

int TrackView::handle(int event) {
   switch(event) {
      case FL_DRAG:
      case FL_PUSH: {
         int b = Fl::event_button();
         int x = Fl::event_x();
         int y = h() - Fl::event_y();
         x += view->getViewRect().left;
         y += view->getViewRect().bottom;
         if (event == FL_PUSH) {
            click_x = x;
            click_y = y;
            drag_x = Fl::event_x();
            drag_y = Fl::event_y();
         }
         switch (tool) {
            case kScrollTool:
               if (b == FL_LEFT_MOUSE && event == FL_DRAG) {
                  view->scrollBy(drag_x - Fl::event_x(), Fl::event_y() - drag_y);
                  boundScroll();
                  updateScrollBars();
               }
               break;
            case kDrawTool:
               if (b == FL_LEFT_MOUSE || b == FL_RIGHT_MOUSE) {
                  int col = x / tiles->getTileWidth();
                  int row = y / tiles->getTileHeight();
                  tiles->getTileAtPoint(x,y) = b == FL_LEFT_MOUSE ? (col % 5) + 5*(row % 5) : kWallTile;
                  ui->setChanged(true);
               }
               break;
            case kPathTool:
               if (event == FL_DRAG && !Fl::event_shift())
                  break;
               ui->setChanged(true);
               if (b == FL_LEFT_MOUSE)
                  pathPolygon.push_back(Tuple(x,y)/kScale);
               else if (b == FL_RIGHT_MOUSE)
                  if (!pathPolygon.empty())
                     pathPolygon.pop_back();
               break;
            case kStartlineTool:
               if (b == FL_LEFT_MOUSE) {
                  ui->setChanged(true);
                  startLine->setLine(Tuple(click_x,click_y),Tuple(x,y));
               }
               break;
         }
         if (event == FL_DRAG) {
            drag_x = Fl::event_x();
            drag_y = Fl::event_y();
         }
         redraw();
         return 1;
      }/*
      case FL_MOUSEWHEEL:
         view->scrollBy(Fl::event_dx(), -Fl::event_dy());
         boundScroll();
         updateScrollBars();
         redraw();
         return 1;*/
      /*
      case FL_RELEASE:
         printf("b: %d x: %d y: %d\n", Fl::event_button(), Fl::event_x(), Fl::event_y());
         redraw();
         return 1;
      case FL_SHORTCUT:
         if (Fl::event_key() == 'x') {
            printf("you pressed x\n");
         }
         redraw();
         return 0;
      */
      default:
         return Fl_Gl_Window::handle(event);
   }
}

void TrackView::importImage(const char *fullFilename)
{
	const int ROAD_COLOR_INDEX = 255;
	Fl_Image	*img;
	unsigned short width, height, bpp, pad_bytes;
	unsigned char value8;
	int i, j;
	unsigned char *cur;
	
	/* Try to load the image file */
	img = loadImage(fullFilename);
	if (img == NULL) {
		ScThrowErr("Error opening image file");
	}
	
	/* calculate some numbers */
	width = img->w();
	height = img->h();
	bpp = img->d();
	/*pad_bytes = (4 - (width%4)) % 4;*/
	//pad_bytes = surface->pitch/bpp - width;
	pad_bytes = 0;
   
	//printf("testing: %d should be equal to %d\n", pad_bytes, (4 - (width%4)) % 4);
	
	if (bpp != 1 && img->count() != 1) {
      printf("bpp: %d count: %d\n", bpp, img->count());
		ScThrowErr("Image must be 256 colours only\n");
	}
	
	newMap(width, height);
   	
	/* Write map tile data to tilemap */
	cur = (unsigned char *)img->data()[0];
	for (i = 0; i < height; i++)
	{
		for (j = 0; j < width; j++)
		{
			value8 = (*cur == ROAD_COLOR_INDEX) ? kWallTile : 5*(i%5) + (j%5);
			(*tiles)(j,i) = value8;
			cur += bpp;
		}
		cur += pad_bytes;
	}
	
	delete img;
	
	printf("loading from image done\n");
   ui->setChanged(false);
   resetScroll();
}

// installs map into user data directory
void TrackView::installMap(const char *name)
{
   char fullname[1024];
   char trackname[256];
   const char kTracklistName[64] = "/tracks/tracklist.xml";
   char *temp;
   
   // first we uninstall the named track
   uninstallMap(name);
   
   // chop off any .xml suffix
   temp = strrchr(name, '.');
   if (temp) {
      if (strcmp(temp, ".xml") == 0)
         temp[0] = '\0';
   }
   // remove path
   temp = strrchr(name, '/');
   if (temp) {
      strncpy(trackname, temp+1, 255);
   }
   
   strncpy(fullname, userDataDir, 1023 - strlen(kTracklistName));
   strcat(fullname, kTracklistName);
   
	// Load the xml file
	TiXmlDocument *xmlDoc = new TiXmlDocument(fullname);
	if (!xmlDoc->LoadFile())
	{
		delete xmlDoc;
		ScThrowErr("xml file not found");
	}
	
	// Add name to track list
	TiXmlHandle *docHandle = new TiXmlHandle(xmlDoc);
	TiXmlElement *xTracklist = docHandle->FirstChildElement("tracklist").Element();
	if (xTracklist == NULL)
		ScThrowErr("xml file could not find tracklist element");
   TiXmlElement track("track");
   track.SetAttribute("name", trackname);
   xTracklist->InsertEndChild(track);
	
	// Save the xml file
	if (!xmlDoc->SaveFile())
	{
		delete xmlDoc;
		ScThrowErr("error saving xml file");
	}
	
	//delete xTracklist;
	delete docHandle;
	delete xmlDoc;
	
   strncpy(fullname, userDataDir, 1023 - 8);
   strcat(fullname, "/tracks/");
   strcat(fullname, trackname);
#ifndef WIN32
   // make sure the tracks directory exists. nb: what whould be done on win32 here???
   char str[1024];
   sprintf(str, "mkdir -p %s", fullname);
   system(str);
#endif
   strcat(fullname, "/");
   strcat(fullname, trackname);
   strcat(fullname, ".xml");
   saveMap(fullname);
   
	printf("install done\n");
}

// uninstall track from user data directory.
// nb: doesn't remove track data itself, only remove entry from tracklist xml file.
void TrackView::uninstallMap(const char *name)
{
   char fullname[1024];
   char trackname[256];
   const char kTracklistName[64] = "/tracks/tracklist.xml";
	const char *attribute;
   char *temp;
   
   // chop off any .xml suffix
   temp = strrchr(name, '.');
   if (temp) {
      if (strcmp(temp, ".xml") == 0)
         temp[0] = '\0';
   }
   // remove path
   temp = strrchr(name, '/');
   if (temp) {
      strncpy(trackname, temp+1, 255);
   }
   
   strncpy(fullname, userDataDir, 1023 - strlen(kTracklistName));
   strcat(fullname, kTracklistName);
   
	// Load the xml file
	TiXmlDocument *xmlDoc = new TiXmlDocument(fullname);
	if (!xmlDoc->LoadFile())
	{
		delete xmlDoc;
		ScThrowErr("xml file not found");
	}
	
	// Remove all occurrences of the track the from track list
	TiXmlHandle *docHandle = new TiXmlHandle(xmlDoc);
	TiXmlElement *xTracklist = docHandle->FirstChildElement("tracklist").Element();
	if (xTracklist == NULL)
		ScThrowErr("xml file could not find tracklist element");
   TiXmlElement *track = xTracklist->FirstChildElement("track");
   TiXmlElement *prevTrack;
	while (track != NULL)
	{
		attribute = track->Attribute("name");
		if (attribute == NULL)
			ScThrowErr("xml track element does not have attribute called \"name\"");
		prevTrack = track;
		track = track->NextSiblingElement("track");
      if (strcmp(attribute, trackname) == 0) {
         xTracklist->RemoveChild(prevTrack);
      }
	}
   
   // Save the xml file
   if (!xmlDoc->SaveFile()) {
      delete xmlDoc;
      ScThrowErr("error saving xml file");
   }
	
	//delete xTracklist;
	delete docHandle;
	delete xmlDoc;
   
	printf("uninstall done\n");
}

void TrackView::saveMap(const char *fullFilename)
{
	char filename[1024] = "";
   char trackname[256] = "";
   char tempStr[256];
   char pathPolygonStr[16384] = "";
   char *suffix;
   Tuple startLineA = startLine->getStartPoint();
   Tuple startLineB = startLine->getEndPoint();
   
	// Setup the xml file
	//printf("xml filename is %s\n", filename);
	TiXmlDocument *xmlDoc = new TiXmlDocument(fullFilename);
	
   TiXmlDeclaration decl("1.0", "UTF-8", "yes");
	TiXmlElement xTrack("track");
	TiXmlElement xTileset("tileset");
	TiXmlElement xTilemap("tilemap");
	TiXmlElement xStartline("startline");
	TiXmlElement xPath("path");
	
	xTileset.InsertEndChild(TiXmlText("default"));
	
	startLineA.x /= tiles->getTileWidth();
	startLineA.y /= tiles->getTileHeight();
	startLineB.x /= tiles->getTileWidth();
	startLineB.y /= tiles->getTileHeight();
	xStartline.SetDoubleAttribute("x1", startLineA.x);
	xStartline.SetDoubleAttribute("y1", startLineA.y);
	xStartline.SetDoubleAttribute("x2", startLineB.x);
	xStartline.SetDoubleAttribute("y2", startLineB.y);
	
   strncpy(filename, fullFilename, 1023);
   suffix = strrchr(filename, '.');
   if (!suffix)
      ScThrowErr("File has no suffix");
   if (strcmp(suffix,".xml") != 0)
      ScThrowErr("File suffix is not .xml");
   suffix[0] = '\0';
   
   suffix = strrchr(filename, '/');
   suffix[0] = '\0';
   if (!suffix)
      ScThrowErr("File path is not absolute");
   strncpy(trackname, &suffix[1], 255);
   
	strcat(trackname, ".map");
	xTilemap.InsertEndChild(TiXmlText(trackname));
	
	for (list<Tuple>::const_iterator i = pathPolygon.begin(); i != pathPolygon.end(); i++) {
		sprintf(tempStr, "%.2lf,%.2lf; ", i->x / tiles->getTileWidth() * kScale, i->y / tiles->getTileWidth() * kScale);
		strcat(pathPolygonStr, tempStr);
	}
	if (!pathPolygon.empty()) {
		pathPolygonStr[strlen(pathPolygonStr) - 2] = '\0';
		xPath.InsertEndChild(TiXmlText(pathPolygonStr));
	}
	else {
		xPath.InsertEndChild(TiXmlText(""));
	}
	
	xTrack.InsertEndChild(xTileset);
	xTrack.InsertEndChild(xTilemap);
	xTrack.InsertEndChild(xStartline);
	xTrack.InsertEndChild(xPath);
	
	strcat(filename, "/");
	strcat(filename, trackname);
   xmlDoc->InsertEndChild(decl);
   xmlDoc->InsertEndChild(xTrack);
	xmlDoc->SaveFile();
	printf("Saving to file %s...\n", filename);
	tiles->saveMap(filename);
	
	// Clear and delete our allocated document and return success ;)
	xmlDoc->Clear();
	delete xmlDoc;
	
	printf("saving done\n");
   ui->setChanged(false);
}

void TrackView::updateScrollBars() {
   const float viewHeight = view->getViewRect().top-view->getViewRect().bottom;
   const float mapHeight = tiles->getBounds().top-tiles->getBounds().bottom;
   const float scrollCentreV = 0.5*(view->getViewRect().bottom+view->getViewRect().top);
   const float scrollBarHeight = viewHeight/mapHeight;
   
   const float viewWidth = view->getViewRect().right-view->getViewRect().left;
   const float mapWidth = tiles->getBounds().right-tiles->getBounds().left;
   const float scrollCentreH = 0.5*(view->getViewRect().right+view->getViewRect().left);
   const float scrollBarWidth = viewWidth/mapWidth;
   
   ui->vertScroll->value((1 - (scrollCentreV - 0.5*viewHeight)/(mapHeight - viewHeight))*(1-scrollBarHeight)*kScrollbarMaxValue, scrollBarHeight*kScrollbarMaxValue, 0, kScrollbarMaxValue);
   ui->horizScroll->value(((scrollCentreH - 0.5*viewWidth)/(mapWidth - viewWidth))*(1-scrollBarWidth)*kScrollbarMaxValue, scrollBarWidth*kScrollbarMaxValue, 0, kScrollbarMaxValue);
}

void TrackView::drawTrack() {
   view->draw();
}

void TrackView::resize(int x, int y, int w, int h)
{
   Fl_Gl_Window::resize(x,y,w,h);
   view->resize(this->w(), this->h());
   boundScroll();
   updateScrollBars();
   redraw();
}

void TrackView::setHorizScrollFrac(double x)
{
   const float width = view->getViewRect().right - view->getViewRect().left;
   view->scrollTo(floor(x*tiles->getBounds().right + 0.5*width), floor(0.5*(view->getViewRect().bottom+view->getViewRect().top)));
   boundScroll();
   redraw();
}

void TrackView::setVertScrollFrac(double x)
{
   const float height = view->getViewRect().top - view->getViewRect().bottom;
   view->scrollTo(floor(0.5*(view->getViewRect().left+view->getViewRect().right)), floor(tiles->getBounds().top - (x*tiles->getBounds().top + 0.5*height)));
   boundScroll();
   redraw();
}

void TrackView::boundScroll()
{
   if (view->getViewRect().left < tiles->getBounds().left)
      view->scrollBy(-view->getViewRect().left, 0);
   if (view->getViewRect().bottom < tiles->getBounds().bottom)
      view->scrollBy(0, -view->getViewRect().bottom);
   if (view->getViewRect().right > tiles->getBounds().right)
      view->scrollBy(tiles->getWidth() - view->getViewRect().right, 0);
   if (view->getViewRect().top > tiles->getBounds().top)
      view->scrollBy(0, tiles->getHeight() - view->getViewRect().top);
}

void TrackView::draw() {
   if (!view)
      init();
    if (!valid()) {
        //glLoadIdentity();
        //glViewport(0,0,w(),h());
        //glOrtho(0, w(), 0, h(), -1, 1);
        //ortho();
        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    }

    glClear(GL_COLOR_BUFFER_BIT);

    glPushMatrix();

    drawTrack();
    
    glPopMatrix();
}
